"""
    Binance Futures ExchangeInterface object.
    Copyright (C) 2022 Matias Kotlik

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import time
from typing import Optional

import binance.exceptions

try:
    from functools import cached_property
except ImportError:  # emerson is "too cool" for py3.8
    from functools import lru_cache

    def cached_property(func):
        return property(lru_cache(maxsize=None)(func))


from datetime import datetime as dt
import pandas as pd
from binance.client import Client
from binance.exceptions import BinanceAPIException

import blankly
from blankly.utils import utils, time_builder
from blankly.enums import MarginType, PositionMode, Side, TimeInForce, HedgeMode, OrderType, ContractType, OrderStatus
from blankly.exchanges.interfaces.futures_exchange_interface import FuturesExchangeInterface
from blankly.exchanges.orders.futures.futures_order import FuturesOrder

BINANCE_FUTURES_FEES = [
    (0.020, 0.040),
    (0.016, 0.040),
    (0.014, 0.035),
    (0.012, 0.032),
    (0.010, 0.030),
    (0.008, 0.027),
    (0.006, 0.025),
    (0.004, 0.022),
    (0.002, 0.020),
    (0.000, 0.017),
]


class BinanceFuturesInterface(FuturesExchangeInterface):
    calls: Client

    @staticmethod
    def to_exchange_symbol(symbol: str):
        return symbol.replace('-', '')

    # TODO binance SPOT should prob be refactored to do this at some point
    def to_blankly_symbol(self, symbol: str):
        return self.symbol_map[symbol]

    @cached_property
    def symbol_map(self):
        symbols = self.calls.futures_exchange_info()['symbols']
        return {
            symbol['symbol']: symbol['baseAsset'] + '-' + symbol['quoteAsset']
            for symbol in symbols
        }

    def __init__(self, exchange_name, authenticated_api):
        super().__init__(exchange_name, authenticated_api)

    def init_exchange(self):
        try:
            self.calls.futures_account()
        except BinanceAPIException as e:
            raise Exception(
                f"{e.error_message}. Are you trying to use your normal exchange keys while in sandbox mode? \nTry "
                "toggling the 'use_sandbox' setting in your settings.json or check if the keys were input "
                "correctly into your keys.json.")
        try:
            # force oneway mode
            self.calls.futures_change_position_mode(dualSidePosition=False)
        except binance.error.ClientError as e:
            if e.error_code != -4059:  # re raise anything other than "already set"
                raise e

    def set_hedge_mode(self, hedge_mode: HedgeMode):
        is_hedge = True if hedge_mode == HedgeMode.HEDGE else False

        if self.calls.futures_get_position_mode(
        )['dualSidePosition'] != is_hedge:
            self.calls.futures_change_position_mode(dualSidePosition=is_hedge)

    def get_hedge_mode(self):
        is_hedge = self.calls.futures_get_position_mode()['dualSidePosition']
        return HedgeMode.HEDGE if is_hedge else HedgeMode.ONEWAY

    def get_leverage(self, symbol: str = None) -> float:
        if not symbol:
            raise Exception(
                'Binance Futures does not support account wide leverage. Use interface.get_leverage(leverage, '
                'symbol) to get leverage for a symbol. ')
        symbol = self.to_exchange_symbol(symbol)
        return float(
            self.calls.futures_position_information(
                symbol=symbol)[0]['leverage'])

    @utils.order_protection
    def set_leverage(self, leverage: int, symbol: str = None):
        if not symbol:
            raise Exception(
                'Binance Futures does not support account wide leverage. Use interface.set_leverage(leverage, '
                'symbol) to set leverage for each symbol you wish to trade. ')
        symbol = self.to_exchange_symbol(symbol)
        return self.calls.futures_change_leverage(
            symbol=symbol, leverage=leverage)['maxNotionalValue']

    @utils.order_protection
    def set_margin_type(self, symbol: str, type: MarginType):
        """
        Set margin type for a symbol.
        """
        symbol = self.to_exchange_symbol(symbol)
        try:
            self.calls.futures_change_margin_type(symbol=symbol,
                                                  marginType=type.upper())
        except BinanceAPIException as e:
            if e.code != -4046:  # -4046 NO_NEED_TO_CHANGE_MARGIN_TYPE (margin type is already set)
                raise e

    def get_margin_type(self, symbol: str):
        # binance doesn't have an API for this
        # an attempt to create a workaround is preserved below
        # proceed at your own risk
        raise NotImplementedError
        # symbol = self.to_exchange_symbol(symbol)
        #
        # # first check for open position
        # position = self.get_positions(symbol)
        # if position:
        #     return position.margin_type
        #
        # # to get the current margin type, try setting CROSSED and see if an exception is raised.
        # try:
        #     self.calls.futures_change_margin_type(
        #         symbol=symbol, marginType=MarginType.CROSSED.upper())
        # except BinanceAPIException as e:
        #     if e.code != -4046:  # -4046 NO_NEED_TO_CHANGE_MARGIN_TYPE (margin type is already set)
        #         return MarginType.CROSSED
        #
        # # if it actually changed, change it back to ISOLATED
        # self.calls.futures_change_margin_type(
        #     symbol=symbol, marginType=MarginType.ISOLATED.upper())
        # return MarginType.ISOLATED

    def get_products(self, filter: str = None) -> dict:
        # https://binance-docs.github.io/apidocs/futures/en/#exchange-information
        res = self.calls.futures_exchange_info()["symbols"]
        products = {}
        for prod in res:
            if '_' in prod['symbol']:
                continue  # don't support expiring contracts
            symbol = prod['baseAsset'] + '-' + prod['quoteAsset']
            products[symbol] = utils.AttributeDict({
                'symbol': symbol,
                'base_asset': prod['baseAsset'],
                'quote_asset': prod['quoteAsset'],
                'contract_type': ContractType.PERPETUAL,
                'price_precision': int(prod['pricePrecision']),
                'size_precision': int(prod['quantityPrecision']),
                'exchange_specific': prod
            })

        if filter:
            return products[filter]
        return products

    def get_account(self, filter=None) -> utils.AttributeDict:
        res = self.calls.futures_account()

        accounts = utils.AttributeDict()
        for asset in res['assets']:
            symbol = asset['asset']
            accounts[symbol] = utils.AttributeDict({
                'available': float(asset['availableBalance']),
                'exchange_specific': asset,
            })

        if filter:
            return accounts[filter]
        return accounts

    def get_positions(self, filter: str = None) -> Optional[dict]:
        account = self.calls.futures_account()

        positions = {}

        # write in data from binance
        for position in account['positions']:
            symbol = position['symbol']
            size = float(position['positionAmt'])
            if size == 0:
                continue  # don't show empty positions
            symbol = self.to_blankly_symbol(symbol)
            margin = MarginType.ISOLATED \
                if position['isolated'] else MarginType.CROSSED
            positions[symbol] = utils.AttributeDict({
                'symbol': symbol,
                'base_asset': utils.get_base_asset(symbol),
                'quote_asset': utils.get_quote_asset(symbol),
                'size': size,
                'position': PositionMode(position['positionSide'].lower()),
                'entry_price': float(position['entryPrice']),
                'contract_type': ContractType.PERPETUAL,
                'leverage': float(position['leverage']),
                'margin_type': margin,
                'unrealized_pnl': float(position['unrealizedProfit']),
                'exchange_specific': position
            })

        if filter:
            return positions.get(filter, None)
        return positions

    @staticmethod
    def to_order_status(status: str):
        if status == 'NEW':
            return OrderStatus.OPEN
        elif status == 'PARTIALLY_FILLED':
            return OrderStatus.PARTIALLY_FILLED
        elif status == 'FILLED':
            return OrderStatus.FILLED
        elif status in ('CANCELED', 'REJECTED'):
            return OrderStatus.CANCELED
        elif status in 'EXPIRED':
            return OrderStatus.EXPIRED
        raise ValueError(f'invalid order status: {status}')

    def parse_order_response(self, response: dict) -> FuturesOrder:
        return FuturesOrder(symbol=self.to_blankly_symbol(response['symbol']),
                            id=int(response['orderId']),
                            status=self.to_order_status(response['status']),
                            size=float(response['origQty']),
                            type=OrderType(response['type'].lower()),
                            contract_type=ContractType.PERPETUAL,
                            side=Side(response['side'].lower()),
                            position=PositionMode(
                                response['positionSide'].lower()),
                            time_in_force=TimeInForce(response['timeInForce']),
                            price=float(response['cumQuote']),
                            limit_price=float(response['price']),
                            response=response,
                            interface=self)

    @utils.order_protection
    def market_order(self,
                     symbol: str,
                     side: Side,
                     size: float,
                     position: PositionMode = PositionMode.BOTH,
                     reduce_only: bool = False) -> FuturesOrder:
        symbol = self.to_exchange_symbol(symbol)
        params = {
            'type': 'MARKET',
            'symbol': symbol,
            'side': side.upper(),
            'quantity': size,
            'positionSide': position.upper(),
            'reduceOnly': reduce_only
        }
        response = self.calls.futures_create_order(**params)

        return self.parse_order_response(response)

    @utils.order_protection
    def limit_order(
            self,
            symbol: str,
            side: Side,
            price: float,
            size: float,
            position: PositionMode = PositionMode.BOTH,
            reduce_only: bool = False,
            time_in_force: TimeInForce = TimeInForce.GTC) -> FuturesOrder:
        symbol = self.to_exchange_symbol(symbol)
        params = {
            'type': 'LIMIT',
            'symbol': symbol,
            'side': side.upper(),
            'price': price,
            'quantity': size,
            'positionSide': position.upper(),
            'reduceOnly': reduce_only,
            'timeInForce': time_in_force.value
        }
        response = self.calls.futures_create_order(**params)

        return self.parse_order_response(response)

    def take_profit(
            self,
            symbol: str,
            side: Side,
            price: float,
            size: float,
            position: PositionMode = PositionMode.BOTH,
            time_in_force: TimeInForce = TimeInForce.GTC) -> FuturesOrder:
        symbol = self.to_exchange_symbol(symbol)
        params = {
            'type': 'TAKE_PROFIT',
            'symbol': symbol,
            'side': side.upper(),
            'price': price,
            'stopPrice': price,  # TODO allow set stop price
            'quantity': size,
            'positionSide': position.upper(),
            'timeInForce': time_in_force.value
        }
        response = self.calls.futures_create_order(**params)

        return self.parse_order_response(response)

    @utils.order_protection
    def stop_loss(
            self,
            symbol: str,
            side: Side,
            price: float,
            size: float,
            position: PositionMode = PositionMode.BOTH,
            time_in_force: TimeInForce = TimeInForce.GTC) -> FuturesOrder:
        symbol = self.to_exchange_symbol(symbol)
        params = {
            'type': 'STOP',
            'symbol': symbol,
            'side': side.upper(),
            'price': price,
            'stopPrice': price,  # TODO allow set stop price
            'quantity': size,
            'positionSide': position.upper(),
            'timeInForce': time_in_force.value
        }
        response = self.calls.futures_create_order(**params)

        return self.parse_order_response(response)

    @utils.order_protection
    def cancel_order(self, symbol: str, order_id: int) -> FuturesOrder:
        symbol = self.to_exchange_symbol(symbol)
        # this library method is broken for some reason 2021-02-25
        res = self.calls.futures_cancel_order(symbol=symbol, orderId=order_id)
        return self.parse_order_response(res)

    def get_open_orders(self, symbol: str = None) -> list:
        if symbol:
            symbol = self.to_exchange_symbol(symbol)
            orders = self.calls.futures_get_open_orders(symbol=symbol)
        else:
            orders = self.calls.futures_get_open_orders()
        return [self.parse_order_response(order) for order in orders]

    def get_order(self, symbol: str, order_id: int) -> FuturesOrder:
        symbol = self.to_exchange_symbol(symbol)
        res = self.calls.futures_get_order(symbol=symbol, orderId=order_id)
        return self.parse_order_response(res)

    def get_price(self, symbol: str) -> float:
        symbol = self.to_exchange_symbol(symbol)
        return float(self.calls.futures_mark_price(symbol=symbol)['markPrice'])

    def get_fees(self) -> utils.AttributeDict:
        # https://www.binance.com/en/blog/futures/trade-crypto-futures-how-much-does-it-cost-421499824684902239
        tier = int(self.calls.futures_account()['feeTier'])
        maker, taker = BINANCE_FUTURES_FEES[tier]
        return utils.AttributeDict({'maker': maker, 'taker': taker})

    @property
    def account(self) -> utils.AttributeDict:
        return self.get_account()

    @property
    def orders(self) -> list:
        return self.get_open_orders()

    def get_product_history(self, symbol, epoch_start, epoch_stop,
                            resolution) -> pd.DataFrame:
        resolution = blankly.time_builder.time_interval_to_seconds(resolution)

        epoch_start = int(utils.convert_epochs(epoch_start))
        epoch_stop = int(utils.convert_epochs(epoch_stop))

        # TODO dedup this from (non-futures) binance interface
        granularities = {
            60: "1m",
            180: "3m",
            300: "5m",
            900: "15m",
            1800: "30m",
            3600: "1h",
            7200: "2h",
            14400: "4h",
            21600: "6h",
            28800: "8h",
            43200: "12h",
            86400: "1d",
            259200: "3d",
            604800: "1w",
            2592000: "1M"
        }

        if resolution not in granularities:
            utils.info_print(
                "Granularity is not an accepted granularity...rounding to nearest valid value."
            )
            resolution = min(granularities,
                             key=lambda gran: abs(resolution - gran))
        gran_string = granularities[resolution]

        # Figure out how many points are needed
        need = int((epoch_stop - epoch_start) / resolution)
        initial_need = need
        window_open = epoch_start
        history = []

        # Convert coin id to binance coin
        symbol = self.to_exchange_symbol(symbol)
        while need > 1000:
            # Close is always 300 points ahead
            window_close = int(window_open + 1000 * resolution)
            history = history + self.calls.futures_klines(
                symbol=symbol,
                startTime=window_open * 1000,
                endTime=window_close * 1000,
                interval=gran_string,
                limit=1000)

            window_open = window_close
            need -= 1000
            time.sleep(.2)
            utils.update_progress((initial_need - need) / initial_need)

        # Fill the remainder
        history_block = history + self.calls.futures_klines(
            symbol=symbol,
            startTime=window_open * 1000,
            endTime=epoch_stop * 1000,
            interval=gran_string,
            limit=1000)

        data_frame = pd.DataFrame(history_block,
                                  columns=[
                                      'time', 'open', 'high', 'low', 'close',
                                      'volume', 'close time',
                                      'quote asset volume', 'number of trades',
                                      'taker buy base asset volume',
                                      'taker buy quote asset volume', 'ignore'
                                  ],
                                  dtype=None)
        # Clear the ignore column, why is that there binance?
        del data_frame['ignore']

        # Want them in this order: ['time (epoch)', 'low', 'high', 'open', 'close', 'volume']

        # Time is so big it has to be cast separately for windows
        data_frame['time'] = data_frame['time'].div(1000).astype(int)

        # Cast dataframe
        data_frame = data_frame.astype({
            'open': float,
            'high': float,
            'low': float,
            'close': float,
            'volume': float,
            'close time': int,
            'quote asset volume': float,
            'number of trades': int,
            'taker buy base asset volume': float,
            'taker buy quote asset volume': float
        })

        # Convert time to seconds
        return data_frame.reindex(
            columns=['time', 'low', 'high', 'open', 'close', 'volume'])

    def get_funding_rate_history(self, symbol: str, epoch_start: int,
                                 epoch_stop: int) -> list:
        symbol = self.to_exchange_symbol(symbol)
        limit = 1000
        history = []
        window_start = epoch_start
        window_end = epoch_stop

        # UNCOMMENT FOR WALRUS
        # while response := self.calls.futures_funding_rate(
        #         symbol=symbol,
        #         startTime=window_start * 1000,
        #         endTime=window_end * 1000,
        #         limit=limit):

        # WARNING! non-walrus code ahead:
        response = True
        while response:
            response = self.calls.futures_funding_rate(
                symbol=symbol,
                startTime=window_start * 1000,
                endTime=window_end * 1000,
                limit=limit)
            # very stinky ^^

            history.extend({
                'rate': float(e['fundingRate']),
                'time': e['fundingTime'] // 1000
            } for e in response)

            if history:
                window_start = history[-1]['time'] + 1
                window_end = min(dt.now().timestamp(), epoch_stop)

        return history

    def get_funding_rate_resolution(self) -> int:
        return time_builder.build_hour() * 8  # 8 hours



import requests
import blankly


def price_event(price, symbol, state: blankly.StrategyState):
    """ This function will give an updated price every 15 seconds from our definition below """
    state.variables['history'].append(price)
    rsi = blankly.indicators.rsi(state.variables['history'])
    if rsi[-1] < 30 and not state.variables['owns_position']:
        # Dollar cost average buy
        buy = int(state.interface.cash / price)
        state.interface.market_order(symbol, side='buy', size=buy)
        state.variables['owns_position'] = True
    elif rsi[-1] > 70 and state.variables['owns_position']:
        # Dollar cost average sell
        curr_value = state.interface.account[state.base_asset].available
        state.interface.market_order(symbol, side='sell', size=int(curr_value))
        state.variables['owns_position'] = False


def init(symbol, state: blankly.StrategyState):
    # Download price data to give context to the algo
    state.variables['history'] = state.interface.history(symbol, to=150, return_as='deque',
                                                         resolution=state.resolution)['close']
    state.variables['owns_position'] = False


if __name__ == "__main__":
    # This downloads an example CSV
    data = requests.get('https://firebasestorage.googleapis.com/v0/b/blankly-6ada5.appspot.com/o/price_examples.csv?'
                        'alt=media&token=3f3c37ee-a87b-46b0-b145-077c0e25254b').text
    with open('./price_examples.csv', 'w') as file:
        file.write(data)

    # Run on the keyless exchange, starting at 100k
    exchange = blankly.KeylessExchange('./price_examples.csv', initial_account_values={'USD': 100000})

    # Use our strategy helper
    strategy = blankly.Strategy(exchange)

    # Make the price event function above run every day
    strategy.add_price_event(price_event, symbol='BTC-USD', resolution='1d', init=init)

    # Backtest the strategy
    results = strategy.backtest(start_date=1568547200, end_date=1642982400)
    print(results)



"""
    FTX websocket definitions
    Copyright (C) 2021 Blankly Finance

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import json
import traceback
import blankly
from blankly.utils.utils import epoch_from_iso8601, to_blankly_symbol
import blankly.exchanges.interfaces.ftx.ftx_websocket_utils as websocket_utils
from blankly.exchanges.interfaces.websocket import Websocket
from blankly.utils.utils import info_print


class Tickers(Websocket):
    def __init__(self, symbol, stream, log=None,
                 pre_event_callback=None, initially_stopped=False, websocket_url="wss://ftx.com/ws/", **kwargs):
        """
        Create and initialize the ticker
        Args:
            symbol: Currency to initialize on such as "BTC-USD"
            log: Fill this with a path to a log file that should be created
            websocket_url: Default websocket URL feed.
        """
        self.__logging_callback, self.__interface_callback, log_message = websocket_utils.switch_type(stream)

        super().__init__(symbol, stream, log, log_message, websocket_url, pre_event_callback, kwargs)

        self.__pre_event_callback_filled = False

        # Start the websocket
        if not initially_stopped:
            self.start_websocket(
                self.on_open,
                self.on_message,
                self.on_error,
                self.on_close,
                self.read_websocket
            )

    def read_websocket(self):
        """
        This is the target that runs
        """
        self.ws.run_forever()

    def on_message(self, ws, message):
        """
        Behavior for this exchange
        """
        received_dict = json.loads(message)
        if received_dict['type'] == 'subscribed':
            info_print(f"Subscribed to {received_dict['channel']}")
            return

        if (self.pre_event_callback is not None) and (not self.__pre_event_callback_filled) and \
                (self.stream == "orderbook"):
            if received_dict['type'] == 'partial':
                try:
                    self.pre_event_callback(received_dict)
                except Exception:
                    traceback.print_exc()

            self.__pre_event_callback_filled = True
            return

        if self.stream == "orderbook":
            interface_response = self.__interface_callback(received_dict['data'])
            self.most_recent_time = interface_response["time"]
            self.time_feed.append(self.most_recent_time)
            self.most_recent_tick = interface_response

            try:
                interface_response['symbol'] = received_dict['market']
                for i in self.callbacks:
                    i(interface_response, **self.kwargs)
            except Exception as e:
                info_print(e)
                traceback.print_exc()
        elif self.stream == "trades":
            for received in received_dict['data']:
                interface_response = self.__interface_callback(received)
                # This could be passed into the received var above which could be cleaner
                interface_response['symbol'] = to_blankly_symbol(received_dict['market'], 'ftx')
                self.ticker_feed.append(interface_response)

                self.most_recent_time = epoch_from_iso8601(received["time"])
                self.time_feed.append(self.most_recent_time)
                self.most_recent_tick = received

                self.log_response(self.__logging_callback, received)

                try:
                    for i in self.callbacks:
                        i(interface_response)
                except Exception as e:
                    info_print(e)
                    traceback.print_exc()

        self.message_count += 1

    def on_error(self, ws, error):
        info_print(error)

    def on_close(self, ws):
        # This repeats the close behavior just in case something happens
        pass

    def on_open(self, ws):
        request = json.dumps({
            "op": "subscribe",
            "channel": self.stream,
            "market": self.symbol
        })
        ws.send(request)

    def restart_ticker(self):
        self.start_websocket(
            self.on_open,
            self.on_message,
            self.on_error,
            self.on_close,
            self.read_websocket
        )



import enum


class MarginType(str, enum.Enum):
    CROSSED = 'crossed'
    ISOLATED = 'isolated'


class Side(str, enum.Enum):
    BUY = 'buy'
    SELL = 'sell'


class PositionMode(str, enum.Enum):
    BOTH = 'both'
    LONG = 'long'
    SHORT = 'short'


class TimeInForce(str, enum.Enum):
    GTC = 'GTC'  # Good Till Cancelled
    FOK = 'FOK'  # Fill or Kill
    IOC = 'IOC'  # Immediate or Cancel


class HedgeMode(str, enum.Enum):
    HEDGE = 'hedge'
    ONEWAY = 'oneway'


class OrderStatus(str, enum.Enum):
    OPEN = 'open'
    PARTIALLY_FILLED = 'partially_filled'
    FILLED = 'filled'
    CANCELED = 'canceled'
    EXPIRED = 'expired'


class OrderType(str, enum.Enum):
    MARKET = 'market'
    LIMIT = 'limit'
    STOP = 'stop'
    TAKE_PROFIT = 'take_profit'
    LIQUIDATION = 'liquidation'


class ContractType(str, enum.Enum):
    PERPETUAL = 'perpetual'
    EXPIRING = 'expiring'
    MOVE = 'move'



"""
    Base Order class
    Copyright (C) 2022 Matias Kotlik

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
from dataclasses import dataclass, field

from blankly.enums import OrderStatus, OrderType, Side, PositionMode, ContractType, TimeInForce
from blankly.exchanges.interfaces.abc_base_exchange_interface import ABCBaseExchangeInterface


@dataclass
class FuturesOrder:
    symbol: str
    id: int
    size: float
    status: OrderStatus
    type: OrderType
    contract_type: ContractType
    side: Side
    position: PositionMode
    price: float
    limit_price: float
    time_in_force: TimeInForce
    response: dict = field(compare=False)
    interface: ABCBaseExchangeInterface = field(compare=False)

    def __str__(self):
        return f"""Order Parameters:
Status: {self.status}
ID: {self.id}
Symbol: {self.symbol}
Type: {self.type}
Contract Type: {self.contract_type}
Side: {self.side}
Size: {self.size}
Position: {self.position}
Price: {self.price}
Time In Force: {self.time_in_force}"""



"""
    Utils file for assisting with trades or market analysis.
    Copyright (C) 2021  Emerson Dove
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.
    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import math

import blankly

import datetime
import json
import sys
import decimal
import os
from datetime import datetime as dt, timezone
from math import ceil, trunc as math_trunc
from typing import Union

import dateutil.parser as dp
import numpy as np
import pandas as pd
import pandas_market_calendars as mcal

from blankly.utils.time_builder import time_interval_to_seconds

# Copy of settings to compare defaults vs overrides
default_general_settings = {
    "settings": {
        "use_sandbox_websockets": False,
        "websocket_buffer_size": 10000,
        "test_connectivity_on_auth": True,

        "coinbase_pro": {
            "cash": "USD"
        },
        "binance": {
            "cash": "USDT",
            "binance_tld": "us"
        },
        "binance_futures": {
            "cash": "USDT",
            "margin_type": "USDT-M"
        },
        "alpaca": {
            "websocket_stream": "iex",
            "cash": "USD",
            "enable_shorting": True,
            "use_yfinance": False,
        },
        "oanda": {
            "cash": "USD"
        },
        "okx": {
            "cash": "USDT"
        },
        "keyless": {
            "cash": "USD"
        },
        "kucoin": {
            "cash": "USDT"
        },
        "ftx": {
            "cash": "USD",
            "ftx_tld": "us"
        },
        "ftx_futures": {
            "cash": "USD",
            "ftx_tld": "com"
        }
    }
}

default_backtest_settings = {
    "price_data": {
        "assets": []
    },
    "settings": {
        "use_price": "close",
        "smooth_prices": False,
        "GUI_output": True,
        "show_tickers_with_zero_delta": False,
        "save_initial_account_value": True,
        "show_progress_during_backtest": True,
        "cache_location": "./price_caches",
        "continuous_caching": True,
        "resample_account_value_for_metrics": "1d",
        "quote_account_value_in": "USD",
        "ignore_user_exceptions": True,
        "risk_free_return_rate": 0.0,
        "benchmark_symbol": None
    }
}

default_notify_settings = {
  "email": {
    "port": 465,
    "smtp_server": "smtp.website.com",
    "sender_email": "email_attached_to_smtp_account@web.com",
    "receiver_email": "email_to_send_to@web.com",
    "password": "my_password"
  },
  "text": {
    "phone_number": "1234567683",
    "provider": "verizon"
  }
}

default_deploy_settings = {
    "main_script": "./bot.py",
    "python_version": '3.7',
    "requirements": "./requirements.txt",
    "working_directory": ".",
    "ignore_files": ['price_caches'],
    "backtest_args": {
        'to': '1y'
    },
    "screener": {
        "schedule": "30 14 * * 1-5"  # Set this default stock-like schedule
    }
}


def load_json_file(override_path=None):
    f = open(override_path, )
    json_file = json.load(f)
    f.close()
    return json_file


class __BlanklySettings:
    def __init__(self, default_path: str, default_settings: dict, not_found_err: str, allow_nonexistent: bool = False):
        """
        Create a class that can manage caching for loading and writing to user preferences with a low overhead.
        This can dramatically accelerate instantiation of new interfaces or other objects
        Args:
            default_path: The default path to look for settings - such as ./settings.json
            default_settings: The default settings in which to compare the loaded settings to. This helps the user
             learn if they're missing important settings and avoids keyerrors later on
            not_found_err: A string that is shown if the file they specify is not found
            allow_nonexistent: Enable this to just get the defaults if the file isn't found
        """
        self.__settings_cache = {}
        self.__default_path = default_path
        self.__default_settings = default_settings
        self.__not_found_err = not_found_err
        self.__allow_nonexistent = allow_nonexistent

    # Recursively check if the user has all the preferences, inform when defaults are missing
    def __compare_dicts(self, default_settings, user_settings):
        for k, v in default_settings.items():
            if isinstance(v, dict):
                if k in user_settings:
                    self.__compare_dicts(v, user_settings[k])
                else:
                    warning_string = "\"" + str(k) + "\" not specified in preferences, defaulting to: \"" + str(v) + \
                                     "\""
                    info_print(warning_string)
                    user_settings[k] = v
            else:
                if k in user_settings:
                    continue
                else:
                    warning_string = "\"" + str(k) + "\" not specified in preferences, defaulting to: \"" + str(v) + \
                                     "\""
                    info_print(warning_string)
                    user_settings[k] = v
        return user_settings

    def load(self, override_path=None) -> dict:
        # Make overrides sticky by changing how the default is set. This is cool because the user can put settings in
        # obscure locations and then continue to load from them
        if override_path is not None:
            self.__default_path = override_path

        if self.__default_path not in self.__settings_cache:
            try:
                preferences = load_json_file(self.__default_path)
            except FileNotFoundError:
                if self.__allow_nonexistent:
                    return self.__default_settings
                    # self.write(self.__default_settings)
                    # # Recursively run this
                    # return self.load(override_path)
                else:
                    raise FileNotFoundError(self.__not_found_err)
            preferences = self.__compare_dicts(self.__default_settings, preferences)
            self.__settings_cache[self.__default_path] = preferences
            return preferences
        else:
            return self.__settings_cache[self.__default_path]

    def write(self, json_information: dict, override_path: str = None):
        # In this case on writes it doesn't change the default path if its overridden globally
        if override_path is None:
            self.__settings_cache[self.__default_path] = json_information
            f = open(self.__default_path, "w")
        else:
            self.__settings_cache[override_path] = json_information
            f = open(override_path, "w")
        f.write(json.dumps(json_information, indent=2))


general_settings = __BlanklySettings('./settings.json', default_general_settings,
                                     "Make sure a settings.json file is placed in the same folder as the project "
                                     "working directory!")

backtest_settings = __BlanklySettings('./backtest.json', default_backtest_settings,
                                      "To perform a backtest, make sure a backtest.json file is placed in the same "
                                      "folder as the project working directory!")

notify_settings = __BlanklySettings('./notify.json', default_notify_settings,
                                    "To send emails locally, make sure a notify.json file is placed in the same folder "
                                    "as the project working directory. This is not necessary when deployed live on "
                                    "blankly cloud.")

deployment_settings = __BlanklySettings('./blankly.json', default_deploy_settings,
                                        "Make sure a blankly.json file is placed in the same folder as the project "
                                        "working directory!", allow_nonexistent=True)


def load_user_preferences(override_path=None) -> dict:
    return general_settings.load(override_path)


def load_backtest_preferences(override_path=None) -> dict:
    return backtest_settings.load(override_path)


def load_deployment_settings() -> dict:
    return deployment_settings.load()


def write_backtest_preferences(json_file, override_path=None):
    backtest_settings.write(json_file, override_path)


def load_notify_preferences(override_path=None) -> dict:
    return notify_settings.load(override_path)


def pretty_print_json(json_object, actually_print=True):
    """
    Json pretty printer for general string usage
    """
    out = json.dumps(json_object, indent=2)
    if actually_print:
        print(out)
    return out


def epoch_from_iso8601(iso8601: str) -> float:
    return dp.parse(iso8601).timestamp()


def convert_input_to_epoch(value: Union[str, dt]) -> float:
    if isinstance(value, str):
        return epoch_from_iso8601(value)
    elif isinstance(value, dt):
        return value.timestamp()
    elif isinstance(value, float):
        return value
    raise ValueError("Incorrect value input given, expected string or value but got: {}".format(type(value)))


def iso8601_from_epoch(epoch) -> str:
    return dt.utcfromtimestamp(epoch).isoformat() + 'Z'

# Removed due to sklearn dependency
# def get_price_derivative(ticker, point_number):
#     """
#     Performs regression n points back
#     """
#     feed = np.array(ticker.get_ticker_feed()).reshape(-1, 1)
#     times = np.array(ticker.get_time_feed()).reshape(-1, 1)
#     if point_number > len(feed):
#         point_number = len(feed)
#
#     feed = feed[-point_number:]
#     times = times[-point_number:]
#     prices = []
#     for i in range(point_number):
#         prices.append(feed[i][0]["price"])
#     prices = np.array(prices).reshape(-1, 1)
#
#     regressor = LinearRegression()
#     regressor.fit(times, prices)
#     regressor.predict(times)
#     return regressor.coef_[0][0]


# def fit_parabola(ticker, point_number):
#     """
#     Fit simple parabola
#     """
#     feed = ticker.get_ticker_feed()
#     times = ticker.get_time_feed()
#     if point_number > len(feed):
#         point_number = len(feed)
#
#     feed = feed[-point_number:]
#     times = times[-point_number:]
#     prices = []
#     for i in range(point_number):
#         prices.append(float(feed[i]["price"]))
#         times[i] = float(times[i])
#
#     # Pull the times back to x=0 so we can know what happens next
#     latest_time = times[-1]
#     for i in range(len(prices)):
#         times[i] = times[i] - latest_time
#
#     return np.polyfit(times, prices, 2, full=True)


def to_blankly_symbol(symbol, exchange, quote_guess=None) -> str:
    if exchange == "binance":
        if quote_guess is not None:
            index = int(symbol.find(quote_guess))
            symbol = symbol[0:index]
            return symbol + "-" + quote_guess
        else:
            # Try your best to try to parse anyway
            quotes = ['BNB', 'BTC', 'TRX', 'XRP', 'ETH', 'USDT', 'USD', 'BUSD', 'AUD', 'BRL', 'EUR', 'GBP', 'RUB',
                      'TRY', 'TUSD', 'USDC', 'PAX', 'BIDR', 'DAI', 'IDRT', 'UAH', 'NGN', 'VAI', 'BVND']
            for i in quotes:
                if __check_ending(symbol, i):
                    return to_blankly_symbol(symbol, 'binance', quote_guess=i)
            raise LookupError("Unable to parse binance coin id of: " + str(symbol))

    if exchange == "coinbase_pro":
        return symbol

    if exchange == "ftx":
        return symbol.replace("/", "-")


def __check_ending(full_string, checked_ending) -> bool:
    check_length = len(checked_ending)
    return checked_ending == full_string[-check_length:]


def to_exchange_symbol(blankly_symbol, exchange):
    if exchange == "binance":
        return blankly_symbol.replace('-', '')
    if exchange == "alpaca":
        return get_base_asset(blankly_symbol)
    if exchange == "coinbase_pro" or exchange == "kucoin":
        return blankly_symbol
    if exchange == 'ftx':
        return blankly_symbol.replace("-", "/")


def get_base_asset(symbol):
    # Gets the BTC of the BTC-USD
    return symbol.split('-')[0]


def get_quote_asset(symbol):
    # Gets the USD of the BTC-USD
    split = symbol.split('-')
    if len(split) > 1:
        return split[1]
    else:
        # This could go wrong
        return 'USD'


def rename_to(keys_array, renaming_dictionary):
    """
    Args:
        keys_array: A two dimensional array that contains information on which keys are changed:
            keys_array = [
                ["key1", "new name"],
                ["id", "user_id"],
                ["frankie", "gerald"],
            ]
        renaming_dictionary: Dictionary to perform the renaming on
    """
    mutated_dictionary = {**renaming_dictionary}
    if "exchange_specific" not in mutated_dictionary:
        mutated_dictionary["exchange_specific"] = {}

    for i in keys_array:
        if i[1] in renaming_dictionary:
            # If we're here this key has already been defined, push it to the specific
            mutated_dictionary["exchange_specific"][i[1]] = renaming_dictionary[i[1]]
            # Then continue with adding the key that we were going to add anyway
        mutated_dictionary[i[1]] = renaming_dictionary[(i[0])]
        del mutated_dictionary[i[0]]
    return mutated_dictionary


# Non-recursive check
def isolate_specific(needed, compare_dictionary):
    """
    This is the parsing algorithm used to homogenize the dictionaries
    """
    # Make a copy of the dictionary, so we don't modify it if you loop over
    compare_dictionary = {**compare_dictionary}
    # Create a row vector for the keys
    column = [column[0] for column in needed]  # ex: ['currency', 'available', 'hold']
    # Create an area to hold the specific data
    if 'exchange_specific' in compare_dictionary:
        exchange_specific = compare_dictionary['exchange_specific']
        del compare_dictionary['exchange_specific']
    else:
        exchange_specific = {}

    for k, v in compare_dictionary.items():
        required = False
        # Check if the value is one of the keys
        for index, val in enumerate(column):
            required = False
            # If it is, there is a state value for it
            if k == val:
                # Push type to value
                if v is not None:
                    compare_dictionary[k] = needed[index][1](v)
                else:
                    compare_dictionary[k] = v
                required = True
                break
        # Must not be found
        # Append non-necessary to the exchange specific dict
        # There has to be a way to do this without raising a flag value
        if not required:
            exchange_specific[k] = compare_dictionary[k]

    # Now we need to remove the keys that we appended to exchange_specific
    for k, v in exchange_specific.items():
        del compare_dictionary[k]

    # If there exists the exchange specific dict in the compare dictionary
    # This is done because after renaming, if there are naming conflicts they will already have been pushed here
    # Because we pulled out the naming conflicts dictionary we just write this directly
    compare_dictionary["exchange_specific"] = exchange_specific

    return compare_dictionary


def convert_epochs(epoch):
    """
    If an epoch time is very long it means that it includes decimals - generally milliseconds. We like the decimal
    format because its unambiguous
    """
    # Hope nobody is fixing this on Friday, June 11, 2128 at 8:53:20 AM (GMT)
    while epoch > 5000000000:
        epoch = epoch / 10
    return epoch


def compare_dictionaries(dict1, dict2, force_exchange_specific=True) -> bool:
    """
    Compare two output dictionaries to check if they have the same keys (excluding "exchange_specific")
    Args:
        dict1 (dict): First dictionary to compare
        dict2 (dict): Second dictionary to compare
        force_exchange_specific (bool): Check if the the exchange_specific tag is in the dictionary
    Returns:
        bool: Are the non specific tags the same?
    """

    # Make copies of the dictionaries
    dict1 = {**dict1}
    dict2 = {**dict2}

    if force_exchange_specific:
        if "exchange_specific" not in dict1:
            raise KeyError("Exchange specific tag not in: " + str(dict1))

        if "exchange_specific" not in dict2:
            raise KeyError("Exchange specific tag not in: " + str(dict2))

        # Safely remove keys now
        del dict1["exchange_specific"]
        del dict2["exchange_specific"]
    else:
        # Remove the tag if its there anyway
        try:
            del dict1["exchange_specific"]
        except KeyError:
            pass
        try:
            del dict2["exchange_specific"]
        except KeyError:
            pass

    valid_keys = []

    for key, value in dict1.items():
        # Check to make sure that key is in the other dictionary
        if key in dict2:
            # Now are they the same type
            if not isinstance(dict2[key], type(value)):
                # Issue detected
                print("Type of key " + str(dict1[key]) + " in dict1 is " + str(type(dict1[key])) +
                      ", but is " + str(type(dict2[key])) + f" in dict2. The name of the key is {key}.")
                return False
            else:
                # If it's a dictionary, go inside of it
                if isinstance(type(value), dict):
                    print("comparing:")
                    print(value)
                    print(dict2[key])
                    if not compare_dictionaries(value, dict2[key], force_exchange_specific=False):
                        return False

                valid_keys.append(key)
                # This code can more specifically compare the key pairs
                # if dict1[key] == dict2[key]:
                #     valid_keys.append(key)
                # else:
                #     print("Types of keys are the same, however " + str(dict1[key]) + " != " + str(dict2[key]) + ".")
                #     return False
        else:
            # Issue detected
            print(key + " found in dict1 but not in " + str(dict2))
            return False

    # Delete these keys so we can check for differences
    for i in valid_keys:
        del dict1[i]
        del dict2[i]

    if dict1 == {} and dict2 == {}:
        return True
    else:
        print("Differing keys:")
        print(dict1)
        print(dict2)
        return False


def update_progress(progress):
    # From this great post: https://stackoverflow.com/a/15860757/8087739
    # update_progress() : Displays or updates a console progress bar
    # Accepts a float between 0 and 1. Any int will be converted to a float.
    # A value under 0 represents a 'halt'.
    # A value at 1 or bigger represents 100%
    bar_length = 10  # Modify this to change the length of the progress bar
    status = ""
    if isinstance(progress, int):
        progress = float(progress)
    if not isinstance(progress, float):
        progress = 0
        status = "error: progress var must be float\r\n"
    if progress < 0:
        progress = 0
        status = "Halt...\r\n"
    if progress >= 1:
        progress = 1
        status = "Done...\r\n"
    block = int(round(bar_length * progress))
    text = "\rProgress: [{0}] {1}% {2}".format("#" * block + "-" * (bar_length - block), round(progress * 100, 2),
                                               status)
    sys.stdout.write(text)
    sys.stdout.flush()


def split_df(df, n):
    return df.groupby(np.arange(len(df)) // n)


def get_ohlcv(candles, n, from_zero: bool):
    if len(candles) < n:
        raise ValueError("Not enough candles provided, required at least {} candles, "
                         "but only received {}".format(n, len(candles)))
    new_candles = pd.DataFrame()
    df = split_df(candles, n)
    new_candles['high'] = df['high'].max().reset_index(drop=True)
    new_candles['low'] = df['low'].min().reset_index(drop=True)
    new_candles['volume'] = df['volume'].sum().reset_index(drop=True)
    new_candles['volume'] = new_candles['volume'].apply(lambda x: float(x))
    if from_zero:
        new_candles['close'] = candles['close'].iloc[0::n].reset_index(drop=True)
        new_candles['open'] = candles['open'].iloc[0::n].reset_index(drop=True)
        new_candles['time'] = candles['time'].iloc[0::n].reset_index(drop=True).astype('int64')
        new_candles['time'] = new_candles['time'].apply(lambda x: np.int64(x))
    else:
        new_candles['close'] = candles['close'].iloc[::n].reset_index(drop=True)
        new_candles['open'] = candles['open'].iloc[::n].reset_index(drop=True)
        new_candles['time'] = candles.index.to_series().iloc[::n].reset_index(drop=True)
        new_candles['time'] = new_candles['time'].apply(lambda x: np.int64(x.timestamp()))
    return new_candles


def aggregate_candles(history: pd.DataFrame, aggregation_size: int):
    """
    Aggregate history data (such as turn 1m data into 15m data)
    Args:
        history: A blankly generated dataframe
        aggregation_size: How many rows of history to aggregate - ex: aggregation_size=15 on 1m data produces
         15m intervals
    """
    aggregated = pd.DataFrame()
    splits = split_df(history, aggregation_size)
    for i in splits:
        # Build candles from the split dataframes
        frame: dict = i[1].to_dict()
        try:
            tick = {
                'time': list(frame['time'].values())[0],
                'open': list(frame['open'].values())[0],
                'high': max(frame['high'].values()),
                'low': min(frame['low'].values()),
                'close': list(frame['close'].values())[-1],
                'volume': sum(frame['volume'].values()),
            }
        except IndexError:
            continue
        aggregated = aggregated.append(tick, ignore_index=True)
    return aggregated


def get_ohlcv_from_list(tick_list: list, last_price: float):
    """
    Created with the purpose of parsing ticker data into a viable OHLCV pattern. The
    Args:
        tick_list (list): List of data containing price ticks. Needs to be at least: [{'price': 343, 'size': 3.4}, ...]
            The data must also be ordered oldest to most recent at the end
        last_price (float): The last price in case there isn't any valid data
    """
    out = {
        'open': last_price,
        'high': last_price,
        'low': last_price,
        'close': last_price,
        'volume': 0
    }
    if len(tick_list) > 0:
        out['open'] = tick_list[0]['price']
        out['close'] = tick_list[-1]['price']

    for i in tick_list:
        out['volume'] = out['volume'] + i['size']

        if i['price'] > out['high']:
            out['high'] = i['price']
        elif i['price'] < out['low']:
            out['low'] = i['price']

    return out


def ceil_date(date, **kwargs):
    secs = datetime.timedelta(**kwargs).total_seconds()
    return dt.fromtimestamp(date.timestamp() + secs - date.timestamp() % secs)


OVERESTIMATE_CONSTANT = 1.5


def get_estimated_start_from_limit(limit, end_epoch, resolution_str, resolution_multiplier):

    nyse = mcal.get_calendar('NYSE')
    required_length = ceil(limit * OVERESTIMATE_CONSTANT)
    resolution = time_interval_to_seconds(resolution_str)

    if resolution == 60 and limit < (1440 / resolution_multiplier):
        return end_epoch - 4 * 86400  # worst case is three day weekend at 9:30am open

    if resolution == 3600 and limit < (24 / resolution_multiplier):
        return end_epoch - 4 * 86400  # worst case is three day weekend at 9:30am open

    temp_start = end_epoch - limit * resolution * resolution_multiplier
    end_date = dt.fromtimestamp(end_epoch, tz=timezone.utc)
    start_date = dt.fromtimestamp(temp_start, tz=timezone.utc)

    schedule = nyse.schedule(start_date=start_date, end_date=end_date)
    date_range = mcal.date_range(schedule, frequency='1D')

    count = 1
    while len(date_range) < required_length:
        temp_start -= 3600 * OVERESTIMATE_CONSTANT * count
        start_date = dt.fromtimestamp(temp_start)
        schedule = nyse.schedule(start_date=start_date, end_date=end_date)
        date_range = mcal.date_range(schedule, frequency='1D')
        count += 1

    return temp_start


class AttributeDict(dict):
    """
    This is adds functions to the dictionary class, no other modifications. This gives dictionaries abilities like:
    print(account.BTC) -> {'available': 1, 'hold': 0}
    account.BTC = "cool"
    print(account.BTC) -> cool
    Basically you can get and set attributes with a dot instead of [] - like dict.available rather than
     dict['available']
    """
    def __getattr__(self, attr):
        # Try catch is wrapped to support copying objects
        try:
            return self[attr]
        except KeyError:
            raise AttributeError(attr)

    def __setattr__(self, attr, value):
        self[attr] = value


def format_with_new_line(original_string, *components):
    for i in components:
        original_string += str(i)

    original_string += '\n'

    return original_string


def trunc(number: float, decimals: int) -> float:
    """
    Truncate a number instead of rounding (ex: trunc(9.9999999, 2) == 9.99 instead of round(9.9999999, 2) == 10.0)
    Args:
        number (float): Number to truncate
        decimals (int): Number of decimals to keep: trunc(9.9999999, 2) == 9.99
    """
    stepper = 10.0 ** decimals
    return math_trunc(stepper * number) / stepper


def info_print(message):
    """
    This prints directly to stderr which allows a way to distinguish package info calls/errors from generic stdout
    Args:
        message: The message to print. INFO: will be prepended
    """
    print('INFO: ' + str(message), file=sys.stderr)


class Email:
    """
    Object wrapper for simplifying interaction with SMTP servers & the blankly.reporter.email function.
    Alternatively a notify.json can be created which automatically integrates with blankly.reporter.email()
    """
    def __init__(self, smtp_server: str, sender_email: str, password: str, port: int = 465):
        """
        Create the email wrapper:
        Args:
            smtp_server: The address of the smtp server
            sender_email: The email attached to the smtp account
            password: The password to log into SMTP
            port: SMTP port - sometimes will fail if not 465
        """
        self.__server = smtp_server
        self.__sender_email = sender_email
        self.__password = password
        self.__port = port

    def send(self, receiver_email: str, message: str):
        """
        Send an email to the receiver_email specified
        Args:
            receiver_email (str): The email that the message is sent to
            message (str): The body of the message
        """
        blankly.reporter.email(email_str=message, smtp_server=self.__server, sender_email=self.__sender_email,
                               receiver_email=receiver_email, password=self.__password, port=self.__port)


def count_decimals(number) -> int:
    """
    Count the number of decimals in a given float: 1.4335 -> 4 or 3 -> 0
    """
    return abs(decimal.Decimal(str(number)).as_tuple().exponent)


def check_backtesting() -> bool:
    """
    Tests if the environment is configured for backtesting. Primarily used for platform deployments but is
    applicable elsewhere
    """
    backtesting = os.getenv('BACKTESTING')

    # Could be undefined
    if backtesting is not None:
        return backtesting == '1'
    else:
        return False


def enforce_base_asset(func):
    """
    Used for get_account functions, this enforces that the user is always getting the base asset which is probably
    what the user meant
    """
    def wrapper(self, symbol=None):
        # Get the base asset if it was defined
        if symbol is not None:
            symbol = get_base_asset(symbol)
        return func(self, symbol=symbol)
    return wrapper


def order_protection(func):
    """
    Decorator to provide protection against live orders inside backtest environment
    """
    def wrapper(*args, **kwargs):
        if blankly._backtesting:
            raise Exception("Blocked attempt at live order inside backtesting environment")
        return func(*args, **kwargs)
    return wrapper


def add_all_products(nonzero_products: dict, all_products: list):
    base_symbols = []
    quote_symbols = []
    for i in all_products:
        base_symbols.append(get_base_asset(i['symbol']))
        quote_symbols.append(get_quote_asset(i['symbol']))

    for i in (base_symbols + quote_symbols):
        if i not in nonzero_products:
            nonzero_products[i] = {
                'available': 0.0,
                'hold': 0.0
            }

    return nonzero_products


def increment_to_precision(increment: float) -> int:
    # quick maths
    # 0.0001 -> 4
    # 0.025 -> 1
    # 0.25 -> 0
    return math.floor(-math.log10(increment))


def trim_df_time_column(df, epoch_start: [int, float], epoch_stop: [int, float]):
    df = df[df['time'] >= epoch_start]
    df = df[df['time'] <= epoch_stop]

    return df




"""
    Exchange metadata for use in the CLI
    Copyright (C) 2022 Matias Kotlik

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from typing import Dict, List
import alpaca_trade_api
from binance.client import Client as BinanceClient
from questionary import Choice

from blankly.deployment.ui import print_failure
from blankly.exchanges.interfaces.alpaca import alpaca_api
from blankly.exchanges.interfaces.coinbase_pro.coinbase_pro_api import API as CoinbaseProAPI
from blankly.exchanges.interfaces.ftx.ftx_api import FTXAPI
from blankly.exchanges.interfaces.oanda.oanda_api import OandaAPI


class Exchange:
    name: str
    symbols: List[str]
    python_class: str
    tlds: List[str]
    display_name: str
    key_info: Dict[str, str]

    def __init__(self, name: str, symbols: List[str], test_func, key_info: List[str] = None, python_class: str = None,
                 tlds: List[str] = None, display_name: str = None, currency: str = 'USD'):
        self.name = name
        self.symbols = symbols
        self.test_func = test_func
        self.key_info = {k.replace('_', ' ').title().replace('Api', 'API'): k  # autogen key instructions
                         for k in key_info or ['API_KEY', 'API_SECRET']}  # default to just key/secret
        self.python_class = python_class or name.replace('_', ' ').title().replace(' ', '')  # snake case to pascalcase
        self.tlds = tlds or []
        self.display_name = display_name or name.replace('_', ' ').title()  # prettify
        self.currency = currency


def kucoin_test_func(auth, tld):
    try:
        from kucoin import client as KucoinAPI
    except ImportError:
        print()  # we are running loading bar at this point, go next line to avoid making a mess
        print_failure('kucoin-python must be installed to check Kucoin API Keys')
        print_failure('Skipping check')
    else:
        return KucoinAPI.User(auth['API_KEY'], auth['API_SECRET'], auth['API_PASS'], auth['sandbox']).get_base_fee


EXCHANGES = [
    Exchange('alpaca', ['MSFT', 'GME', 'AAPL'],
             lambda auth, tld: alpaca_trade_api.REST(key_id=auth['API_KEY'],
                                                     secret_key=auth['API_SECRET'],
                                                     base_url=(alpaca_api.live_url,
                                                               alpaca_api.paper_url)[auth['sandbox']]
                                                     ).get_account()),
    Exchange('binance', ['BTC-USDT', 'ETH-USDT', 'SOL-USDT'],
             lambda auth, tld: BinanceClient(api_key=auth['API_KEY'], api_secret=auth['API_SECRET'],
                                             tld=tld, testnet=auth['sandbox']).get_account(),
             tlds=['com', 'us'], currency='USDT'),
    Exchange('coinbase_pro', ['BTC-USD', 'ETH-USD', 'SOL-USD'],
             lambda auth, tld: CoinbaseProAPI(api_key=auth['API_KEY'], api_secret=auth['API_SECRET'],
                                              api_pass=auth['API_PASS'],
                                              API_URL=('https://api.pro.coinbase.com/',
                                                       'https://api-public.sandbox.pro.coinbase.com/')[auth['sandbox']]
                                              ).get_accounts(),
             key_info=['API_KEY', 'API_SECRET', 'API_PASS']),
    Exchange('ftx', ['BTC-USD', 'ETH-USD', 'SOL-USD'],
             lambda auth, tld: FTXAPI(auth['API_KEY'], auth['API_SECRET'], tld).get_account_info(),
             tlds=['com', 'us'], python_class='FTX', display_name='FTX'),
    Exchange('oanda', ['BTC-USD', 'ETH-USD', 'SOL-USD'],
             lambda auth, tld: OandaAPI(personal_access_token=auth['PERSONAL_ACCESS_TOKEN'],
                                        account_id=auth['ACCOUNT_ID'], sandbox=auth['sandbox']).get_account(),
             key_info=['ACCOUNT_ID', 'PERSONAL_ACCESS_TOKEN']),
    Exchange('kucoin', ['BTC-USDT', 'ETH-USDT', 'SOL-USDT'], kucoin_test_func,
             key_info=['API_KEY', 'API_SECRET', 'API_PASS'], currency='USDT'),
]

EXCHANGE_CHOICES_NO_KEYLESS = [Choice(exchange.display_name, exchange)
                               for exchange in EXCHANGES]
EXCHANGE_CHOICES = EXCHANGE_CHOICES_NO_KEYLESS[:]
EXCHANGE_CHOICES.append(Choice('Keyless/No Exchange', False))


def exc_display_name(name):
    return next((exchange.display_name for exchange in EXCHANGES if name == exchange.name), name)



from blankly.enums import OrderStatus, Side, ContractType, OrderType, PositionMode, TimeInForce
from blankly.exchanges.interfaces.binance_futures.binance_futures import BinanceFutures
from blankly.exchanges.interfaces.ftx_futures.ftx_futures import FTXFutures
from blankly.exchanges.futures.futures_exchange import FuturesExchange
from blankly.exchanges.interfaces.futures_exchange_interface import FuturesExchangeInterface
from blankly.exchanges.orders.futures.futures_order import FuturesOrder
from blankly.utils import utils

import time
import functools
import os

import pytest
from _pytest.python import Metafunc
from _pytest.python_api import approx

from contextlib import contextmanager

FUTURES_EXCHANGES = [
    BinanceFutures(keys_path="./tests/config/keys.json",
                   preferences_path="./tests/config/settings.json",
                   portfolio_name="Futures Test Key"),
    # FTXFutures(keys_path="./tests/config/keys.json",
    #            preferences_path="./tests/config/settings.json",
    #            portfolio_name="Dotcom Test Account"),
]
SYMBOLS = {
    'ftx_futures': ['SOL-USD', 'ETH-USD'],
    'binance_futures': ['BCH-USDT', 'ETH-USDT']
}


def homogeneity_testing(func=None, check_values: bool = False):
    # allow using without arguments
    # this trick brought to you by Python Cookbook
    if func is None:
        return functools.partial(homogeneity_testing, values=check_values)

    results = {}

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        test_name = os.environ.get('PYTEST_CURRENT_TEST')
        test_name = test_name[test_name.find('[') + 1:test_name.find(']')]
        result = func(*args, **kwargs)
        compare_results(results, result, check_values)
        results[test_name] = result
        return result

    return wrapper


def compare_results(previous_outputs: dict, result, check_values: bool):
    for prev_name, prev_results in previous_outputs.items():
        compare_values(prev_name, prev_results, result, check_values)


def compare_values(other_name, other, this, check_values: bool):
    assert type(this) == type(
        other
    ), f'comparing to {other_name}: types do not match. this={this} other={other}'
    # gapped by python 3.7
    # ), f'comparing to {other_name}: types do not match. {this=} {other=}'
    if isinstance(this, list):
        if this and other:
            for val in this[1:]:  # check for self-homogenity
                compare_values('self', val, this[0], check_values)
            for val in other:
                compare_values(other_name, val, this[0], check_values)
    if isinstance(other, dict):
        other.pop('exchange_specific', None)
        this.pop('exchange_specific', None)
        assert this.keys() == other.keys(
        ), f'comparing to {other_name}: dict keys do not match. ' \
           f'symmetric difference: {set(this.keys()).symmetric_difference(other.keys())}'
        for key in other:
            compare_values(other_name, other[key], this[key], check_values)
    elif check_values:
        assert this == other, f'comparing to {other_name}: values are not equal.'


def wait_till_filled(interface: FuturesExchangeInterface, order: FuturesOrder):
    retries = 0
    res = interface.get_order(order.symbol, order.id)
    while res.status != OrderStatus.FILLED:
        if retries > 2:
            pytest.fail(f"order was not filled in time. status: {res.status}")
        time.sleep(1 << retries)
        retries += 1
        res = interface.get_order(res.symbol, order.id)
    return res


def place_order(interface: FuturesExchangeInterface, symbol: str, side: Side, funds: int, reduce_only: bool = False):
    product = interface.get_products(symbol)
    price = interface.get_price(symbol)
    order_size = utils.trunc(funds / price, product.size_precision)

    order = interface.market_order(symbol, side, order_size, reduce_only=reduce_only)
    assert order.symbol == symbol
    assert 0 <= order.id
    assert order.status in (OrderStatus.OPEN, OrderStatus.FILLED)
    assert order.type == OrderType.MARKET
    assert order.contract_type == ContractType.PERPETUAL
    assert order.side == side
    assert order.limit_price == 0
    assert order.position == PositionMode.BOTH
    assert order.time_in_force == TimeInForce.GTC

    res = wait_till_filled(interface, order)
    assert res.status == OrderStatus.FILLED

    if reduce_only:
        # 1% tolerance
        assert res.price <= price * order_size * 1.01
        assert res.size <= order_size * 1.01

        # set size equal so we can compare
        order.size = res.size
    else:
        assert res.price == approx(price * order_size, rel=0.01)
        assert res.size == approx(order_size)

    # set price, status equal so we can compare
    order.price = res.price
    order.status = res.status
    assert order == res

    return order


def sell(interface: FuturesExchangeInterface, symbol: str, funds: int = 20, reduce_only: bool = False):
    return place_order(interface, symbol, Side.SELL, funds, reduce_only)


def buy(interface: FuturesExchangeInterface, symbol: str, funds: int = 20, reduce_only: bool = False):
    return place_order(interface, symbol, Side.BUY, funds, reduce_only)

def close_all(futures_interface: FuturesExchangeInterface):
    for symbol in futures_interface.get_positions():
        close_position(futures_interface, symbol)

    for order in futures_interface.get_open_orders():
        futures_interface.cancel_order(order.symbol, order.id)
    assert len(futures_interface.get_positions()) == 0


def close_position(interface: FuturesExchangeInterface, symbol: str):
    """
    Exit position
    Args:
        interface: the interface to sell on
        symbol: the symbol to sell
    """
    position = interface.get_positions(symbol)
    if not position:
        return
    if position.size < 0:
        order = interface.market_order(symbol,
                                       Side.BUY,
                                       position.size * -2,
                                       reduce_only=True)
    elif position.size > 0:
        order = interface.market_order(symbol,
                                       Side.SELL,
                                       position.size * 2,
                                       reduce_only=True)
    else:
        pytest.fail('position size is zero')
    wait_till_filled(interface, order)
    assert interface.get_positions(symbol) is None


@contextmanager
def cancelling_order(interface: FuturesExchangeInterface, symbol: str):
    """
    Context manager for placing and cancelling a limit order from within a test.
    Useful for testing paths that require active orders.
    Args:
        interface: The interface to place an order on
        symbol: The symbol to buy

    Returns: The order object as returned from the interface
    """
    product = interface.get_products(symbol)
    price = interface.get_price(symbol)

    # place our limit order
    order_size = utils.trunc(100 / price, product.size_precision)  # $1 order
    limit_price = utils.trunc(price * 0.95, product.price_precision)
    order = interface.limit_order(symbol, Side.BUY, limit_price, order_size)
    assert order.symbol == symbol
    assert 0 <= order.id
    assert order.size == approx(order_size)
    assert order.status == OrderStatus.OPEN
    assert order.type == OrderType.LIMIT
    assert order.contract_type == ContractType.PERPETUAL
    assert order.side == Side.BUY
    assert order.position == PositionMode.BOTH
    assert order.limit_price == approx(limit_price)
    assert order.time_in_force == TimeInForce.GTC

    yield order

    res = interface.cancel_order(symbol, order.id)
    assert res.status == OrderStatus.CANCELED

    assert order != res
    res.status = OrderStatus.OPEN
    assert order == res



"""
    Local trading definition.
    Copyright (C) 2021  Emerson Dove

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import copy

import blankly.utils.utils as utils
from blankly.utils.exceptions import InvalidOrder


class LocalAccount:
    def __init__(self, currencies: dict) -> None:
        """
        Create the local account to paper trade with.

        Args:
            currencies: (dict) with key/value pairs such as {'BTC': {'available: 2.3, 'hold': 0}}
        """
        # This is used for shorting. It largely corresponds with margin
        self.__granted_value = {}
        self.local_account = utils.AttributeDict(currencies)

    def override_local_account(self, currencies: dict) -> None:
        """
        After initialization, this is a setter for overriding the internal values
        """
        self.local_account = currencies

    def trade_local(self, symbol, side, base_delta, quote_delta, quote_resolution, base_resolution) -> None:
        """
        Trade on the local & static account

        Args:
            symbol (string): Pair such as 'BTC-USD'
            side (string): Purchase side such as 'buy' or 'sell' (currently unused but required)
            base_delta (float): A number specifying the change in base currency, such as -2.4 BTC by selling or
             +1.2 BTC by buying
            quote_delta (float): Similar to the base_delta - a number specifying the change in quote currency, such as
             -20.12 USD for buying or +103.21 USD for selling
            quote_resolution (int): The number of decimals that are supported for the quote
            base_resolution (int): The number of decimals that are supported for the base
        """

        # Extract the base and quote pairs of the currency
        base = utils.get_base_asset(symbol)
        quote = utils.get_quote_asset(symbol)

        # Push these abstracted deltas to the local account
        try:
            self.local_account[base]['available'] = utils.trunc(self.local_account[base]['available'] + base_delta,
                                                                base_resolution)
        except KeyError:
            raise KeyError("Base currency specified not found in local account")

        try:
            self.local_account[quote]['available'] = utils.trunc(self.local_account[quote]['available'] + quote_delta,
                                                                 quote_resolution)
        except KeyError:
            raise KeyError("Quote currency specified not found in local account")

    def test_trade(self, currency_pair, side, qty, quote_price, quote_resolution, base_resolution, shortable) -> bool:
        """
        Test a paper trade to see if you have the funds

        Args:
            currency_pair (string): Pair such as 'BTC-USD'
            side (string): Purchase side such as 'buy' or 'sell'
            qty (float): Amount to buy of the base currency ex: (2.3 BTC of BTC-USD)
            quote_price (float): Price of the base currency in the currency pair - (1 BTC is valued at 40,245 in
             BTC-USD)
        """
        # Nobody knows what's happening if it's shorting
        if shortable:
            base_asset = utils.get_base_asset(currency_pair)
            quote_asset = utils.get_quote_asset(currency_pair)
            base_account = self.local_account[base_asset]
            quote_account = self.local_account[quote_asset]

            # Initialize a granted value if not already created
            if quote_asset not in self.__granted_value:
                self.__granted_value[quote_asset] = 0

            if side == "sell":
                # Selling is the only thing that can give granted value
                target_quantity = base_account['available'] - qty

                # If they sell and its still just positive that's a valid condition
                if target_quantity > 0:
                    # print(self.__granted_value)
                    return True

                # Remember in this case if it crosses the zero line and becomes negative some positive
                #  value doesn't convert into negative, so we have to understand how much deficit we need
                if base_account['available'] > 0:
                    # This just the part of the quantity request that goes over zero
                    # Qty should be bigger in this case, so we're looking for that difference
                    size_increase_negative = qty - base_account['available']
                else:
                    # In this case because we're already negative we're increasing by this size
                    size_increase_negative = qty

                # In this case the target funds is less than zero which means we have to grant value (margin)
                negative_deficit = self.__granted_value[quote_asset] + abs(size_increase_negative * quote_price)
                margin_requested = utils.trunc(abs(size_increase_negative * quote_price), quote_resolution)
                if negative_deficit >= quote_account['available']:
                    raise InvalidOrder(f"Not enough margin to perform short - total margin available: "
                                       f"{quote_account['available']}, margin already granted: "
                                       f" {self.__granted_value[quote_asset]}, margin requested: "
                                       f"{margin_requested}")
                else:
                    # Add this on margin and allow the order
                    self.__granted_value[quote_asset] += margin_requested
                    # print(self.__granted_value)
                    return True

            elif side == "buy":
                # Buying is the only thing that can eat granted value
                target_size = base_account['available'] + qty
                requested_funds = utils.trunc(abs(qty * quote_price), quote_resolution)
                target_funds = quote_account['available'] - requested_funds

                if target_funds < 0:
                    raise InvalidOrder(f'Not enough funds to buy - available: '
                                       f'{quote_account["available"]}, requested: '
                                       f'{requested_funds}.')

                # The valid condition is if the current amount and the final amount are both greater than zero
                if (target_size >= 0) and (base_account['available'] >= 0):
                    pass
                    # print("Eating 0")
                    # print(self.__granted_value)

                if (target_size >= 0) and (base_account['available'] <= 0):
                    # This will eat a small portion of the granted value (margin)
                    self.__granted_value[quote_asset] -= utils.trunc(abs(base_account['available'] *
                                                                         quote_price), quote_resolution)
                    # print(f"Eating {utils.trunc(abs(base_account['available'] * quote_price), quote_resolution)}")
                    # print(self.__granted_value)

                if (target_size <= 0) and (base_account['available'] <= 0):
                    # This will also eat the qty increase
                    self.__granted_value[quote_asset] -= utils.trunc(abs(qty * quote_price), quote_resolution)
                    # print(f"Eating {utils.trunc(abs(qty * quote_price), quote_resolution)}")
                    # print(self.__granted_value)

                # Correct this just in case they sell for a profit
                if self.__granted_value[quote_asset] < 0:
                    self.__granted_value[quote_asset] = 0

                return True
        else:
            if side == 'buy':
                quote = utils.get_quote_asset(currency_pair)
                account = self.local_account[quote]
                current_funds = self.local_account[quote]['available']
                purchase_funds = utils.trunc(quote_price * qty, quote_resolution)

                # If you have more funds than the purchase requires then return true
                if current_funds >= purchase_funds:
                    return True
                else:
                    raise InvalidOrder("Insufficient funds. Available: " +
                                       str(current_funds) +
                                       " hold: " +
                                       str(account['hold']) +
                                       " requested: " +
                                       str(purchase_funds) + ".")

            elif side == 'sell':
                base = utils.get_base_asset(currency_pair)
                account = self.local_account[base]
                current_base = utils.trunc(account['available'], base_resolution)

                # If you have more base than the sell requires then return true
                if current_base >= qty:
                    return True
                else:
                    raise InvalidOrder("Not enough base currency. Available: " +
                                       str(current_base) +
                                       ". hold: " +
                                       str(account['hold']) +
                                       ". requested: " +
                                       str(qty) + ".")

        raise LookupError("Invalid purchase side")

    def get_accounts(self) -> utils.AttributeDict:
        """
        Get the paper trading local account
        """
        return copy.deepcopy(utils.AttributeDict(self.local_account))

    def get_account(self, asset_id) -> utils.AttributeDict:
        """
        Get a single account under an asset id
        """
        return copy.deepcopy(utils.AttributeDict(self.local_account[asset_id]))

    def update_available(self, asset_id, new_value):
        self.local_account[asset_id]['available'] = new_value

    def update_hold(self, asset_id, new_value):
        self.local_account[asset_id]['hold'] = new_value



"""
    Abstraction for creating interval driven user strategies for trading futures.
    Copyright (C) 2022 Matias Kotlik

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import typing

from blankly.exchanges.futures.futures_exchange import FuturesExchange
from blankly.exchanges.futures.futures_strategy_logger import FuturesStrategyLogger
from blankly.exchanges.interfaces.futures_exchange_interface import FuturesExchangeInterface
from blankly.exchanges.interfaces.paper_trade.backtest_result import BacktestResult
from blankly.frameworks.strategy.strategy_base import StrategyBase


class FuturesStrategy(StrategyBase):
    exchange: FuturesExchange
    interface: FuturesExchangeInterface

    def __init__(self, exchange: FuturesExchange):
        super().__init__(
            exchange, FuturesStrategyLogger(exchange.interface, strategy=self))

    def backtest(self,
                 to: str = None,
                 initial_values: dict = None,
                 start_date: typing.Union[str, float, int] = None,
                 end_date: typing.Union[str, float, int] = None,
                 save: bool = False,
                 settings_path: str = None,
                 callbacks: list = None,
                 **kwargs) -> BacktestResult:
        raise NotImplementedError

    def time(self) -> float:
        """
        Return the current time, or backtesting time if we are backtesting.
        """
        raise NotImplementedError



"""
    Headers to provide a clean UI to users
    Copyright (C) 2021  Emerson Dove

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import abc
import typing
from blankly.data.data_reader import PriceReader, EventReader, TickReader


class ABCBacktestController(abc.ABC):
    # Only becomes defined when backtest is run
    initial_time = None

    @abc.abstractmethod
    def add_prices(self, symbol: str,
                   resolution: [str, int, float],
                   to: str = None,
                   start_date: typing.Union[str, float, int] = None,
                   stop_date: typing.Union[str, float, int] = None):
        """
        Add prices directly to the backtest engine
        """
        pass

    @abc.abstractmethod
    def add_custom_prices(self, price_reader: PriceReader):
        pass

    @abc.abstractmethod
    def add_custom_events(self, event_reader: EventReader):
        pass

    @abc.abstractmethod
    def add_tick_events(self, tick_reader: TickReader):
        pass

    @abc.abstractmethod
    def value_account(self):
        """
        Value your account at this time
        """
        pass



"""
    Coinbase Pro ticker manager.
    Copyright (C) 2021  Emerson Dove

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import json
import time
import traceback

import blankly
import blankly.exchanges.interfaces.coinbase_pro.coinbase_pro_websocket_utils as websocket_utils
from blankly.exchanges.interfaces.websocket import Websocket
from blankly.utils.utils import info_print


def create_ticker_connection(id, url, channel):
    ws = None  # create_connection(url, sslopt={"cert_reqs": ssl.CERT_NONE})
    request = """{
    "type": "subscribe",
    "product_ids": [
        \"""" + id + """\"
    ],
    "channels": [ \"""" + channel + """\" ]
    }"""
    ws.send(request)
    return ws


# This could be needed:
# "channels": [
#     {
#         "name": "ticker",
#         "product_ids": [
#             \"""" + id + """\"
#         ]
#     }
# ]


class Tickers(Websocket):
    def __init__(self, symbol, stream, log=None,
                 pre_event_callback=None, initially_stopped=False, websocket_url="wss://ws-feed.pro.coinbase.com",
                 **kwargs):
        """
        Create and initialize the ticker
        Args:
            symbol: Currency to initialize on such as "BTC-USD"
            log: Fill this with a path to a log file that should be created
            websocket_url: Default websocket URL feed.
        """
        self.__logging_callback, self.__interface_callback, log_message = websocket_utils.switch_type(stream)

        super().__init__(symbol, stream, log, log_message, websocket_url, pre_event_callback, kwargs)

        self.__pre_event_callback_filled = False

        # Start the websocket
        if not initially_stopped:
            self.start_websocket(
                self.on_open,
                self.on_message,
                self.on_error,
                self.on_close,
                self.run_forever
            )

    def run_forever(self):
        """
        This is the target from the super

        It's renamed to run_forever in coinbase pro because I think that ancient code below is cool
        """
        self.ws.run_forever()

    """
    This function has some of the oldest code in the entire package
    """
    def read_websocket(self):
        # This is unique because coinbase first sends the entire orderbook to use
        if self.__pre_event_callback is not None and self.__stream == "level2":
            received_string = json.loads(self.ws.recv())
            if received_string['type'] == 'snapshot':
                try:
                    self.__pre_event_callback(received_string)
                except Exception:
                    traceback.print_exc()

        counter = 0
        # TODO port this to "WebSocketApp" found in the websockets documentation
        while self.ws.connected:
            # In case the user closes while its reading from the websocket, this will let it expire
            persist_connected = self.ws.connected
            try:
                received_string = self.ws.recv()
                received = json.loads(received_string)
                # Modify time to use epoch
                self.__most_recent_time = blankly.utils.epoch_from_iso8601(received["time"])
                received["time"] = self.__most_recent_time
                self.__time_feed.append(self.__most_recent_time)

                if self.__log:
                    if counter % 100 == 0:
                        self.__file.close()
                        self.__file = open(self.__filePath, 'a')
                    line = self.__logging_callback(received)
                    self.__file.write(line)

                # Manage price events and fire for each manager attached
                interface_message = self.__interface_callback(received)
                self.__ticker_feed.append(interface_message)
                self.__most_recent_tick = interface_message

                try:
                    for i in self.__callbacks:
                        i(interface_message)
                except Exception:
                    traceback.print_exc()

                counter += 1
            except Exception as e:
                if persist_connected:
                    traceback.print_exc()
                    pass
                else:
                    traceback.print_exc()
                    print("Error reading ticker websocket for " + self.__id + " on " +
                          self.__stream + ": attempting to re-initialize")
                    # Give a delay so this doesn't eat up from the main thread if it takes many tries to initialize
                    time.sleep(2)
                    self.ws.close()
                    # Update response
                    self.response = self.ws.recv()

    def on_message(self, ws, message):
        received_string = message
        received = json.loads(received_string)

        if received['type'] == 'subscriptions':
            info_print(f"Subscribed to {received['channels']}")
            return

        # This is unique because coinbase first sends the entire orderbook to use
        if (self.pre_event_callback is not None) and (not self.__pre_event_callback_filled) and \
                (self.stream == "level2"):
            if received['type'] == 'snapshot':
                try:
                    self.pre_event_callback(received)
                except Exception:
                    traceback.print_exc()

            self.__pre_event_callback_filled = True
            return

        # Modify time to use epoch
        self.most_recent_time = blankly.utils.epoch_from_iso8601(received["time"])
        received["time"] = self.most_recent_time
        self.time_feed.append(self.most_recent_time)

        self.log_response(self.__logging_callback, received)

        # Manage price events and fire for each manager attached
        interface_message = self.__interface_callback(received)
        self.ticker_feed.append(interface_message)
        self.most_recent_tick = interface_message

        try:
            for i in self.callbacks:
                i(interface_message, **self.kwargs)
        except Exception:
            traceback.print_exc()

        self.message_count += 1

    def on_error(self, ws, error):
        info_print(error)

    def on_close(self, ws):
        # This repeats the close behavior just in case something happens
        pass

    def on_open(self, ws):
        request = json.dumps({
            'type': 'subscribe',
            'product_ids': [self.symbol],
            'channels': [self.stream]
        })
        ws.send(request)

    def restart_ticker(self):
        """
        This is the only abstract function that should be placed individually
        into each websocket file
        """
        self.start_websocket(
            self.on_open,
            self.on_message,
            self.on_error,
            self.on_close,
            self.read_websocket
        )



"""
    An improved strategy simulation layer built for spot and futures trading
    Copyright (C) 2021  Emerson Dove

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import threading
import typing

from blankly.exchanges.interfaces.paper_trade.backtest_controller import BackTestController, BacktestResult
from blankly.exchanges.interfaces.paper_trade.abc_backtest_controller import ABCBacktestController
from blankly.exchanges.interfaces.paper_trade.paper_trade import PaperTrade
from blankly.exchanges.exchange import Exchange
from blankly.utils.time_builder import time_interval_to_seconds
import time


class Model:
    def __init__(self, exchange: Exchange):
        self.__exchange = exchange
        self.__exchange_cache = self.__exchange
        self.is_backtesting = False

        self.interface = exchange.interface

        self.has_data = True

        self.backtester: ABCBacktestController = BackTestController(self)
        # Type these internal calls to the specific backtester
        self.__backtester: BackTestController = self.backtester

    def backtest(self, args, initial_values: dict = None, settings_path: str = None, kwargs=None) -> BacktestResult:
        # Construct the backtest controller
        if not isinstance(self.__exchange, Exchange):
            raise NotImplementedError

        if kwargs is None:
            kwargs = {}

        # Toggle backtesting
        self.is_backtesting = True
        self.__exchange = PaperTrade(self.__exchange)
        self.interface = self.__exchange.interface
        backtest = self.__backtester.run(args,
                                         initial_account_values=initial_values,
                                         exchange=self.__exchange,
                                         backtest_settings_path=settings_path,
                                         **kwargs
                                         )
        self.is_backtesting = False
        self.__exchange = self.__exchange_cache
        self.has_data = True

        return backtest

    def run(self, args: typing.Any = None) -> threading.Thread:
        thread = threading.Thread(target=self.main, args=(args,))
        thread.start()
        return thread

    def main(self, args):
        raise NotImplementedError("Add a main function to your strategy to run the model.")

    def event(self, type_: str, data: any):
        """
        Override this to gain access to any custom events passed in
        """
        pass

    def websocket_update(self, data):
        pass

    @property
    def time(self):
        if not self.is_backtesting:
            return time.time()
        else:
            return self.__backtester.time

    def sleep(self, seconds: [int, float, str]):
        seconds = time_interval_to_seconds(seconds)
        if not self.is_backtesting:
            time.sleep(seconds)
        else:
            self.__backtester.sleep(seconds)



"""
    This class is able to maintain the state of the process and run screeners when needed
    Copyright (C) 2021  Emerson Dove

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import datetime
from datetime import timezone
import time
import runpy
import os
import __main__
import threading
from blankly.utils.utils import info_print, trunc


class ScreenerRunner:
    def __init__(self, cronjob: str):
        try:
            from croniter import croniter
            self.croniter = croniter
        except ImportError:
            raise ImportError("To run screeners locally, please \"pip install croniter\".")
        self.__main = __main__.__file__

        self.croniter = self.croniter(cronjob, datetime.datetime.fromtimestamp(time.time()).astimezone(timezone.utc))

        self.__stopped = False

        # Start the execution thread
        threading.Thread(target=self.execute).start()

    def execute(self):
        while True:
            # Sleep until the next
            next_run_datetime = self.croniter.get_next()
            next_run_in = next_run_datetime - time.time()
            info_print(f"Starting screener...the next screener run will be started in {trunc(next_run_in, 2)} "
                       f"seconds or at {datetime.datetime.fromtimestamp(next_run_datetime)}")
            time.sleep(next_run_in)
            if not self.__stopped:
                # Start the screener in a different thread
                threading.Thread(target=self.__executor).start()
            else:
                break

    def __executor(self):
        # The execute function to run the modules
        main_script_abs = os.path.abspath(self.__main)
        runpy.run_path(main_script_abs, {}, "__main__")

    def stop(self):
        self.__stopped = True



"""
    Functions for managing Exchange API Keys
    Copyright (C) 2022 Matias Kotlik

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import json

import alpaca_trade_api
from binance.client import Client as BinanceClient

from blankly.deployment.exchange_data import Exchange
from blankly.deployment.ui import print_failure, show_spinner
from blankly.exchanges.interfaces.coinbase_pro.coinbase_pro_api import API as CoinbaseProAPI
from blankly.exchanges.interfaces.ftx.ftx_api import FTXAPI
from blankly.exchanges.interfaces.oanda.oanda_api import OandaAPI

from blankly.deployment.ui import confirm


def load_keys():
    try:
        with open('keys.json', 'r') as file:
            return json.load(file)
    except FileNotFoundError:
        return {}


def write_keys(data):
    with open('keys.json', 'w') as file:
        json.dump(data, file, indent=4)


def add_key(exchange: Exchange, tld: str, key_name: str, data: dict):
    # load old keys file
    saved_data = load_keys()

    keys = saved_data.setdefault(exchange.name, {})

    # change name until it's not already in our file
    if not key_name:
        key_name = 'example-portfolio'
    if key_name in keys:
        key_name = key_name + '-1'
    while key_name in keys:
        base, num = key_name.rsplit('-', 1)
        key_name = f'{base}-{int(num) + 1}'

    keys[key_name] = data  # this writes to saved_data

    key_is_valid = check_key(exchange, tld, data)
    if key_is_valid or confirm('Would you like to save this key anyway?', default=False).unsafe_ask():
        write_keys(saved_data)
        return True
    return False


def check_key(exchange: Exchange, tld: str, auth):
    with show_spinner(f'Checking {exchange.display_name} API Key') as spinner:
        try:
            exchange.test_func(auth, tld)
        except Exception:
            spinner.fail(f'Failed to connect to {exchange.display_name}. Check that your keys are valid.')
            return False
        spinner.ok(f'Checked {exchange.display_name} API Key')
        return True



"""
    Utils for creating objects related to interacting with Coinbase Pro.
    Copyright (C) 2021  Emerson Dove

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import uuid


class CoinbaseProUtils:
    def __init__(self):
        pass

    """
    Size: Amount of base currency to buy or sell
    Price: Price per bitcoin
    Ex: Buy .001BTC at $15,000 is generateLimitOrder(.001, 15000, "buy", "BTC-USD")
    """
    """
    Order Place Example:
    order = {
        'size': 1.0,
        'price': 1.0,
        'side': 'buy',
        'product_id': 'BTC-USD',
    }
    (size in currency (like .01 BTC), buy/sell (string), product id (BTC-USD))
    """

    @staticmethod
    def generate_limit_order(size, price, side, product_id):
        order = {
            'size': size,
            'price': price,
            'side': side,
            'product_id': product_id,
        }
        return order

    @staticmethod
    def generate_uuid():
        # This should be a reasonable approximation for order UUIDs
        return uuid.uuid4()



from blankly import Screener, ScreenerState


def evaluator(symbol, state: ScreenerState):
    pass


def formatter(results, state: ScreenerState):
    pass


if __name__ == '__main__':
    exchange = EXCHANGE_CLASS()  # initialize our exchange
    screener = Screener(exchange, evaluator, symbols=[], formatter=formatter)  # find oversold

    print(screener.formatted_results)



"""
    Inherited authentication object
    Copyright (C) 2021  Arun Annamalai, Emerson Dove

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import abc
import json

from blankly.exchanges.auth.utils import load_auth
from blankly.utils.exceptions import AuthError
from blankly.utils.utils import info_print


class AuthConstructor(abc.ABC):
    def __init__(self, keys_file: str, portfolio_name: str, exchange: str, needed_keys: list):
        """
        Create an auth interface
        Args:
            keys_file (str): filepath to keys.json
            portfolio_name (str): name of portfolio
            exchange (str): name of exchange
        """
        # self.portfolio_name = portfolio_name
        self.exchange = exchange

        # Load from file
        self.portfolio_name, self.keys = load_auth(exchange, keys_file, portfolio_name)

        # Create an error message template to throw if needed
        error_message = ""

        # Make a copy of the keys dict & list to avoid modifying passed variables
        keys_dict = {**self.keys}
        needed_keys = list.copy(needed_keys)
        for i in needed_keys:
            if i in keys_dict.keys():
                keys_dict.pop(i)
            else:
                # Append a description header if not present
                example_str = json.dumps({'portfolio': {'API_KEY': '********************',
                                                        'API_SECRET': '********************',
                                                        f'{i}': False}}, indent=2)
                if error_message == "":
                    error_message += f"Error while loading authentication. Required keys for this are missing: \n" \
                                     f"Please add the \"{i}\" key to the keys.json file. For example: \n" \
                                     f"{example_str}"

                if i != "sandbox":
                    raise AuthError(error_message)
                else:
                    info_print(f"Please add the sandbox keys to your keys.json file. The use_sandbox setting will be "
                               f"removed in the next update: \n"
                               f"{example_str}")

        if len(keys_dict.keys()) > 0:
            info_print(f"Additional keys for Exchange: {self.exchange} Portfolio: {self.portfolio_name} will be"
                       f" ignored.")



"""
    Implementation for Kucoin Websockets
    Copyright (C) 2022  Emerson Dove

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import json
import time
import traceback

import blankly.exchanges.interfaces.kucoin.kucoin_websocket_utils as websocket_utils
from blankly.exchanges.interfaces.websocket import Websocket
from blankly.utils.utils import info_print


class Tickers(Websocket):
    def __init__(self, symbol, stream, websocket_url, log=None,
                 pre_event_callback=None, initially_stopped=False,
                 id_=None, **kwargs):
        """
        Create and initialize the ticker
        Args:
            symbol: Currency to initialize on such as "BTC-USD"
            log: Fill this with a path to a log file that should be created
            websocket_url: Default websocket URL feed.
        """
        self.id = id_
        self.__logging_callback, self.__interface_callback, log_message = websocket_utils.switch_type(stream)

        super().__init__(symbol, stream, log, log_message, websocket_url, pre_event_callback, kwargs)

        # Start the websocket
        if not initially_stopped:
            self.start_websocket(
                self.on_open,
                self.on_message,
                self.on_error,
                self.on_close,
                self.read_websocket
            )

    def read_websocket(self):
        # Main thread to sit here and run
        self.ws.run_forever()

    def on_message(self, ws, message):
        """
        Exchange specific actions to perform when receiving a message
        """
        # print(message)
        message = json.loads(message)

        if message['type'] == 'subscribe':
            channel = message['topic'].split(":", 1)[0].split("/", 2)[2]
            info_print(f"Subscribed to {channel}")
            return
        elif message['type'] == 'welcome' or message['type'] == 'ack':
            return

        # parsed_received_trades = websocket_utils.trade_interface(message)

        # for received in parsed_received_trades:
            # ISO8601 is converted to epoch in process_trades
        self.most_recent_time = time.time()
        self.time_feed.append(self.most_recent_time)

        self.log_response(self.__logging_callback, message)

        # Manage price events and fire for each manager attached
        interface_message = self.__interface_callback(message)
        self.ticker_feed.append(interface_message)
        self.most_recent_tick = interface_message

        try:
            for i in self.callbacks:
                i(interface_message, **self.kwargs)
        except Exception as e:
            info_print(e)
            traceback.print_exc()

        self.message_count += 1

    def on_error(self, ws, error):
        info_print(error)

    def on_close(self, ws):
        # This repeats the close behavior just in case something happens
        pass

    def on_open(self, ws):
        request = json.dumps({
            'id': self.id,
            'type': 'subscribe',
            'topic': f'/market/{self.stream}:{self.symbol}',
            'privateChannel': False,
            'response': True
        })
        ws.send(request)

    def restart_ticker(self):
        self.start_websocket(
            self.on_open,
            self.on_message,
            self.on_error,
            self.on_close,
            self.read_websocket
        )



"""
    Formatting for platform backtest result
    Copyright (C) 2021  Emerson Dove

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import uuid


def __dict_from_df_dict(dict_of_df: dict) -> dict:
    """
    This creates a serializable dictionary from a dictionary with pandas df:
    {
        'AAPL': df.DataFrame
        'MSFT': df.DataFrame
    }
    ->
    {
        'AAPL': df.DataFrame.to_dict()
        'MSFT': df.DataFrame.to_dict()
    }
    The function does not check for nested dataframes

    :arg dict_of_df: A dictionary with dataframes as values of the internal keys
    :return: dict
    """
    output_dict = {}
    for i in list(dict_of_df.keys()):
        output_dict[i] = dict_of_df[i].to_dict()

    return output_dict


def __compress_dict_series(values_column: dict, time_dictionary: dict):
    """
    Remove duplicate keys from the account value or anything that is associated with a time series

    :param values_column: A dictionary with index as key and any sort of value
    :param time_dictionary: A dictionary with index also as key but a time that corresponds to the same index in the
        values column
    :return: None
    """
    values_keys = list(values_column.keys())
    time_keys = list(time_dictionary.keys())

    # Check to see if the dictionaries are longer than 0
    if len(values_keys) == 0:
        return {}

    last_value = values_column[values_keys[0]]
    output_dict = {
        time_dictionary[time_keys[0]]: last_value
    }
    for i in range(len(values_column)):
        if last_value != values_column[values_keys[i]]:
            output_dict[time_dictionary[time_keys[i]]] = values_column[values_keys[i]]

            last_value = values_column[values_keys[i]]

    return output_dict


def __parse_backtest_trades(trades: list, limit_executed: list, limit_canceled: list, market_executed: list):
    """
    Determine the lifecycle of limit orders

    :param trades: The list of trades from the backtest
    :param limit_executed: The list of limit orders that were executed
    :param limit_canceled: The list of limit orders that were canceled
    :param market_executed: The list of executed market orders
    :return: None
    """

    # Now just parse if there should be an executed time or a canceled time
    for i in range(len(trades)):
        try:
            trades[i]['time'] = trades[i].pop('created_at')
        except KeyError:
            pass
        if trades[i]['type'] == 'limit':
            # TODO this wastes a few CPU cycles at the moment so it could be cleaned up
            for j in limit_executed:
                if trades[i]['id'] == j['id']:
                    trades[i]['time'] = j['executed_time']
                    break

            for j in limit_canceled:
                if trades[i]['id'] == j['id']:
                    trades[i]['canceledTime'] = j['canceled_time']
                    break
        elif trades[i]['type'] == 'market':
            # This adds in the execution price for the market orders
            trades[i]['type'] = 'spot-market'
            for j in market_executed:
                if trades[i]['id'] == j['id']:
                    trades[i]['price'] = j['executed_price']
                    break

    return trades


def format_metric(metrics: dict, display_name, type_: str):
    return {
        "value": metrics[display_name],
        "display_name": display_name,
        "type": type_
    }


def format_platform_result(backtest_result):
    """
    Export the finished backtest result

    Args:
        backtest_result: A BacktestResult object to export
    """
    # We're given a dictionary of dataframes and that has to be transformed to a dictionary of dictionaries
    history_and_returns = __dict_from_df_dict(backtest_result.history_and_returns)

    # Grab a list of the traded assets
    # The paper trade interface format may change in the future to be more optimized
    traded_symbols = []
    for i in backtest_result.trades['created']:
        if i['symbol'] not in traded_symbols:
            traded_symbols.append(i['symbol'])

    # Set the account values to None
    raw_or_resampled_account_values = None

    # Now grab the account value dictionary itself
    # Now just replicate the format of the resampled version
    # This was the annoying backtest glitch that almost cost us an investor meeting so its important
    history = history_and_returns['history']
    account_value_name = 'Account Value (' + backtest_result.quote_currency + ')'
    raw_or_resampled_account_values = {
        'value': history[account_value_name],
        'time': history['time']
    }

    # Now grab the raw account values
    compressed_asset_column = __compress_dict_series(raw_or_resampled_account_values['value'],
                                                     raw_or_resampled_account_values['time'])

    compressed_array_conversion = []
    for i in compressed_asset_column:
        compressed_array_conversion.append({
            'time': i,
            'value': compressed_asset_column[i]
        })

    compressed_asset_keys = list(compressed_asset_column.keys())
    first_account_value = compressed_asset_column[compressed_asset_keys[0]]
    last_account_value = compressed_asset_column[compressed_asset_keys[-1]]

    trades = __parse_backtest_trades(backtest_result.trades['created'],
                                     backtest_result.trades['limits_executed'],
                                     backtest_result.trades['limits_canceled'],
                                     backtest_result.trades['executed_market_orders'])

    # Now change the metrics keys to be nicer
    metrics = backtest_result.metrics
    refined_metrics = {
        'calmar': format_metric(metrics, 'Calmar Ratio', 'number'),
        'cagr': format_metric(metrics, 'Compound Annual Growth Rate (%)', 'number'),
        'cavr': format_metric(metrics, 'Conditional Value-at-Risk', 'number'),

        'cum_returns': format_metric(metrics, 'Cumulative Returns (%)', 'number'),
        'max_drawdown': format_metric(metrics, 'Max Drawdown (%)', 'number'),
        'resampled_time': format_metric(metrics, 'Resampled Time', 'number'),

        'risk_free_rate': format_metric(metrics, 'Risk Free Return Rate', 'number'),
        'sharpe': format_metric(metrics, 'Sharpe Ratio', 'number'),
        'sortino': format_metric(metrics, 'Sortino Ratio', 'number'),

        'value_at_risk': format_metric(metrics, 'Value-at-Risk', 'number'),
        'variance': format_metric(metrics, 'Variance (%)', 'number'),
        'volatility': format_metric(metrics, 'Volatility', 'number')
    }

    backtest_result.metrics = refined_metrics
    return {
        'exchange': backtest_result.exchange,
        'symbols': traded_symbols,
        'quote_asset': backtest_result.quote_currency,
        'start_time': backtest_result.start_time,
        'stop_time': backtest_result.stop_time,
        'account_values': compressed_array_conversion,
        'trades': trades,
        'metrics': refined_metrics,
        'indicators': {},  # Add to this array when we support
        # strategy indicators
        'user_callbacks': backtest_result.user_callbacks,
        'initial_account_value': first_account_value,
        'final_account_value': last_account_value,
        'backtest_id': str(uuid.uuid4())
        # Everything above here is put into the firebase root
    }



import time
import blankly.utils.utils as utils


def switch_type(stream):
    if stream == "tickers":
        return trade, \
               trade_interface, \
               "time, system_time, price, open_24h, volume_24h, low_24h, high_24h, best_bid, best_ask," \
               "last_traded_price, last_traded_size\n"

    elif stream == "books":
        return no_callback, \
               no_callback, \
               ""
    else:
        return no_callback, no_callback, ""


def no_callback(message):
    return message


def trade(received):
    line = str(received['data']['ts'] / 1000) + "," + str(time.time()) + "," + received['data']['sodUtc0'] + "," + \
           received[
               'data']['open24h'] + "," + received['data']['volCcy24h'] + "," + received['data']['low24h'] + "," + \
           received[
               'data']['high24h'] + "," + received['data']['bidSz'] + "," + received['data']['askSz'] + "," + received[
               'data']['last'] + "," + received['data']['lastSz'] + "\n"
    return line


def trade_interface(message):
    needed = [
        ["symbol", str],
        ["price", float],
        ["time", float],
        ["trade_id", int],
        ["size", float]
    ]

    symbol = message['instId']
    new_symbol = '-'.join(symbol.split('-')[:2])
    message['symbol'] = new_symbol
    message['size'] = message['lastSz']
    message['trade_id'] = None
    message['price'] = message['last']
    message['time'] = message['ts']

    return utils.isolate_specific(needed, message)
